import ast
import argparse
import os
import sys


class Analyzer(ast.NodeVisitor):

  future_division = False

  def __init__(self, file_path, analyze=["division", "round"], silent=False):
    with open(file_path) as f:
      source_code = f.read()
      self.file_path = f.name
    self.tree = ast.parse(source_code)
    self.source_code = source_code.split('\n')
    self.first_div_visit = True
    self.analyze = analyze
    self.silent = silent

  def __call__(self):
    self.visit(self.tree)

  def visit_ImportFrom(self, node):
    if node.module == "__future__" and any(alias.name == "division"
                                           for alias in node.names):
      self.future_division = True
    self.generic_visit(node)

  def print_message(self, node, mark):
    code = self.source_code[node.lineno - 1]
    stripped_code = code.strip()
    leading_spaces = len(code) - len(stripped_code)
    print("%s %d,%d:\n>>> [%c] %s\n%s^" % \
          (self.file_path, node.lineno, node.col_offset, mark, stripped_code,
           " " * (8 + node.col_offset - leading_spaces)))

  def visit_Call(self, node):
    if isinstance(node.func, ast.Name) and node.func.id is 'round' \
       and 'round' in self.analyze and self.get_type(node) == 1:
      self.print_message(node, 'R')
    self.generic_visit(node)

  def visit_AugAssign(self, node):
    self.visit_BinOp(node)

  def visit_BinOp(self, node):
    try:
      if isinstance(node.op, ast.Div) and 'division' in self.analyze:
        if self.first_div_visit:
          self.first_div_visit = False
          if not self.future_division:
            print("No __future__.division in %s\n" % self.file_path)
        if isinstance(node, ast.BinOp):
          a = node.left
          b = node.right
        else:
          a = node.target
          b = node.value
        x = min(self.get_type(a), self.get_type(b))
        if self.future_division and x == 1:
          x = -1
        mark = ''
        if x == -2:
          mark = "$" # redundancy can be simplified
        elif x == 1:
          mark = '!' # must be changed
        elif not x and not self.silent:
          mark = '?' # unknown
        if mark:
          self.print_message(node, mark)
      self.generic_visit(node)
    except Exception as e:
      if not self.silent:
        print("Could not analyze %s %d,%d: %s\n" % \
          (self.file_path, node.lineno, node.col_offset, e))

  # Type analysis. Possible results:
  # - -2: redundant float
  # - -1: float
  # -  0: unknown
  # -  1: int
  def get_type(self, node):
    if isinstance(node, ast.Num):
      if type(node.n) is int:
        return 1
      if type(node.n) is float:
        return -2 if int(node.n) == node.n else -1
      if type(node.n) is complex:
        return -1
      assert False, node
    elif isinstance(node, ast.NameConstant):
      if node.value in (True, False):
        return 1
    elif isinstance(node, ast.Call):
      if isinstance(node.func, ast.Name):
        if node.func.id in ('bool', 'int', 'len'):
          return 1
        if node.func.id == 'float':
          # WARN: cast to float could not be removed automatically if we're not
          #       sure arg is a number
          return -2
        if node.func.id is 'round':
          return 1 if len(node.args) == 1 else -1
        if not node.keywords:
          if node.func.id == 'abs' and len(node.args) == 1:
            return self.get_type(*node.args)
    elif isinstance(node, ast.UnaryOp):
      if isinstance(node.op, (ast.Not, ast.Invert)):
        return 1
      if isinstance(node.op, (ast.UAdd, ast.USub)):
        return self.get_type(node.operand)
      assert False, node
    elif isinstance(node, ast.BinOp):
      if isinstance(node.op, (ast.LShift, ast.RShift, ast.BitOr, ast.BitXor,
                              ast.BitAnd)):
        return 1
      x = min(self.get_type(node.left), self.get_type(node.right))
      if isinstance(node.op, ast.Div) and self.future_division and x == 1:
        return -1
      return x
    elif isinstance(node, ast.BoolOp):
      a = None
      for value in node.values:
        b = self.get_type(value)
        if a is not None:
          if a * b <= 0:
            break
          if a < b:
            continue
        a = b
      else:
        return a
    return 0


def main():
  parser = argparse.ArgumentParser()
  parser.add_argument('path', help="file or directory to analyze")
  parser.add_argument('-A', '--analyze', nargs='+',
                      default=Analyzer.__init__.__defaults__[0],
                      help="list of nodes to analyze")
  parser.add_argument('-S', '--silent', action='store_true',
                      help="do not print unknown cases (\"[?]\") and errors")
  parser.add_argument('-I', '--ignore', nargs='+', default=[],
                      help="list of files or directories to ignore")
  args = parser.parse_args()

  for file_path in [args.path] if os.path.isfile(args.path) else \
                   [os.path.join(dirpath, name)
                    for dirpath, _, filenames in os.walk(args.path)
                    for name in filenames]:
    if any(file_path.startswith(i) for i in args.ignore):
      continue
    if file_path.endswith('.py'):
      try:
        Analyzer(file_path, args.analyze, args.silent)()
      except Exception as e:
        if not args.silent:
          print("Could not analyze %s: %s\n" % (file_path, e))

if __name__ == '__main__':
  main()
